"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
//import * as reob from "./reob"
const Internal_1 = require("./Internal");
const reob_1 = require("./reob");
function getClass(o) {
    if (o)
        return o.constructor;
    else
        return undefined;
}
exports.getClass = getClass;
// it seems that the local variable that "reflect" uses is prone to the same difficulties when it gets loaded
// multiple times. This is why it's been removed until it is supported by the Runtime directly.
/**
 * @hidden
 */
function defineMetadata(propertyName, value, cls) {
    if (!cls.hasOwnProperty("_reobAnnotations")) {
        Internal_1.setNonEnumerableProperty(cls, "_reobAnnotations", {});
    }
    var _reobAnnotations = cls._reobAnnotations;
    _reobAnnotations[propertyName] = value;
}
/**
 * @hidden
 */
function getMetadata(propertyName, cls) {
    if (cls.hasOwnProperty("_reobAnnotations"))
        return cls["_reobAnnotations"] ? cls["_reobAnnotations"][propertyName] : undefined;
    else {
        return undefined;
    }
}
/**
 * Declares a class as an entity. Can be used either with or without given name for the entity.
 * @Decorator
 */
function Entity(name) {
    var entityName;
    if (typeof name == "string") {
        entityName = name;
    }
    else {
        entityName = getClassName(name);
    }
    var f = function (p1) {
        var typeClass = p1;
        defineMetadata("persistence:entity", true, typeClass);
        exports.entityClasses[entityName] = typeClass;
        Object.defineProperty(p1, "_reobClassName", {
            value: entityName,
            writable: false,
            configurable: false,
            enumerable: false
        });
    };
    if (typeof name == "string") {
        return f;
    }
    else {
        f(name);
    }
}
exports.Entity = Entity;
/**
 * Convenience decorator that combines @CollectionUpdate and @Remote.
 * @Decorator
 */
function RemoteCollectionUpdate(t, functionName, objectDescriptor) {
    CollectionUpdate(t, functionName, objectDescriptor);
    Remote(t, functionName, objectDescriptor);
}
exports.RemoteCollectionUpdate = RemoteCollectionUpdate;
/**
 * Used to declare a function of a class as a "collection update". That means that whenever the function is called
 * the same operation is also invoked on the document in the collection.
 * @Decorator
 */
function CollectionUpdate(p1, functionName, desc) {
    var options = {};
    setPropertyProperty(p1, functionName, "collectionUpdate", options);
    var entityClass = p1.constructor;
    monkeyPatch(entityClass.prototype, functionName, function (originalFunction, ...args) {
        var _reobObjectContext = reob_1.SerializationPath.getObjectContext(this);
        if (!_reobObjectContext || !_reobObjectContext.handler || !_reobObjectContext.handler.collectionUpdate) {
            return originalFunction.apply(this, args);
        }
        else {
            return _reobObjectContext.handler.collectionUpdate(entityClass, functionName, this, originalFunction, args, _reobObjectContext.request).then((r) => {
                originalFunction.apply(this, args);
                reob_1.SerializationPath.updateObjectContexts(this, _reobObjectContext.handler, _reobObjectContext.request);
                return r;
            });
        }
    });
    desc.value = entityClass.prototype[functionName];
    return desc;
}
exports.CollectionUpdate = CollectionUpdate;
function ArrayOrMap(typeClassName) {
    let f = function (targetPrototypeObject, propertyName) {
        setPropertyProperty(targetPrototypeObject, propertyName, "type", typeClassName);
        setPropertyProperty(targetPrototypeObject, propertyName, "arrayOrMap", true);
    };
    return f;
}
exports.ArrayOrMap = ArrayOrMap;
function ArrayType(typeClassName) {
    return ArrayOrMap(typeClassName);
}
exports.ArrayType = ArrayType;
function DictionaryType(typeClassName) {
    return ArrayOrMap(typeClassName);
}
exports.DictionaryType = DictionaryType;
/**
 * Used to declare which property is used as the value for "_id" in mongo.
 */
function Id(targetPrototypeObject, propertyName) {
    DocumentName("_id")(targetPrototypeObject, propertyName);
}
exports.Id = Id;
/**
 * In the case where a document consists of multiple subdocuments and the documents are mapped to classes.
 * A property of a class that represents a sub document which is annotated as Parent will hold the reference to the
 * containing document.
 */
function Parent(targetPrototypeObject, propertyName) {
    setPropertyProperty(targetPrototypeObject, propertyName, "parent", 1);
}
exports.Parent = Parent;
/**
 * Declares that a property of an entity is not persisted.
 */
function Ignore(targetPrototypeObject, propertyName) {
    setPropertyProperty(targetPrototypeObject, propertyName, "ignore", true);
}
exports.Ignore = Ignore;
function PrivateToServer(targetPrototypeObject, propertyName) {
    setPropertyProperty(targetPrototypeObject, propertyName, "privateToServer", true);
}
exports.PrivateToServer = PrivateToServer;
function DocumentName(name) {
    return function (targetPrototypeObject, propertyName) {
        var objNames = getMetadata("objectNames", targetPrototypeObject);
        if (!objNames) {
            objNames = {};
            defineMetadata("objectNames", objNames, targetPrototypeObject);
        }
        var documentNames = getMetadata("documentNames", targetPrototypeObject);
        if (!documentNames) {
            documentNames = {};
            defineMetadata("documentNames", documentNames, targetPrototypeObject);
        }
        objNames[name] = propertyName;
        documentNames[propertyName] = name;
    };
}
exports.DocumentName = DocumentName;
function Type(typeClassName) {
    return function (targetPrototypeObject, propertyName) {
        setPropertyProperty(targetPrototypeObject, propertyName, "type", typeClassName);
    };
}
exports.Type = Type;
/**
 * Returns the property previously defined with {@link idProperty} or the _id property.
 * @param o {Object} The object
 * @hidden
 */
function getId(o) {
    if (typeof o == "undefined")
        return undefined;
    var idPropertyName = getIdPropertyName(getClass(o));
    if (!idPropertyName)
        throw new Error("No id property defined for object of class " + getClass(o));
    else
        return o[idPropertyName];
}
exports.getId = getId;
/**
 * @deprecated
 */
function MeteorMethod(p1, p2, d) {
    Remote(p1, p2, d);
}
exports.MeteorMethod = MeteorMethod;
function Remote(p1, p2, d) {
    var f = function (methodOptions, descriptor) {
        setPropertyProperty(methodOptions.parentObject, methodOptions.propertyName, "methodOptions", methodOptions);
        monkeyPatch(methodOptions.parentObject, methodOptions.propertyName, function (originalFunction, ...args) {
            var _reobObjectContext = reob_1.SerializationPath.getObjectContext(this);
            var request = _reobObjectContext ? _reobObjectContext.request : undefined;
            if (!_reobObjectContext || !_reobObjectContext.handler || !_reobObjectContext.handler.webMethod) {
                return originalFunction.apply(this, args);
            }
            else {
                return _reobObjectContext.handler.webMethod(methodOptions.parentObject.constructor, methodOptions.propertyName, this, originalFunction, args, request);
            }
        });
        if (descriptor) {
            descriptor.value = methodOptions.parentObject[methodOptions.propertyName];
            return descriptor;
        }
    };
    // parameter case : options
    if (typeof p1 == "object" && typeof p2 == "undefined") {
        return (prototype, functionName, d) => {
            var options = p1;
            options.parentObject = prototype;
            options.propertyName = functionName;
            return f(options, d);
        };
    }
    else if (typeof p1 == "string" && typeof p2 == "undefined") {
        return (prototype, functionName, d) => {
            var options = {};
            options.parentObject = prototype;
            options.propertyName = functionName;
            options.name = p1;
            return f(options, d);
        };
    }
    else {
        var options = {};
        options.parentObject = p1;
        options.propertyName = p2;
        return f(options, d);
    }
}
exports.Remote = Remote;
function PostCreate(t, functionName, objectDescriptor) {
    setPropertyProperty(t, functionName, "postCreate", true);
}
exports.PostCreate = PostCreate;
function getMethodOptions(typeClass, functionName) {
    return getPropertyProperty(typeClass, functionName, "methodOptions");
}
exports.getMethodOptions = getMethodOptions;
function getClassName(cls) {
    if (!cls) {
        return undefined;
    }
    else if (cls['_reobClassName']) {
        return cls['_reobClassName'];
    }
    else {
        if (cls['name']) {
            return cls['name'];
        }
        else if (cls.constructor && cls.constructor['name']) {
            return cls.constructor['name'];
        }
        else {
            var n = cls.toString();
            n = n.substr('function '.length);
            n = n.substr(0, n.indexOf('('));
            return n;
        }
    }
}
exports.getClassName = getClassName;
function getRemoteFunctionNames(typeClass) {
    return getPropertyNamesOfPropertiesThatHaveAProperty(typeClass, "methodOptions");
}
exports.getRemoteFunctionNames = getRemoteFunctionNames;
function getPostCreateFunctionNames(typeClass) {
    return getPropertyNamesOfPropertiesThatHaveAProperty(typeClass, "postCreate");
}
exports.getPostCreateFunctionNames = getPostCreateFunctionNames;
function getRemoteFunctionNamesByObject(o) {
    var cls = getClass(o);
    return getRemoteFunctionNames(cls);
}
exports.getRemoteFunctionNamesByObject = getRemoteFunctionNamesByObject;
// public static getAllMethodFunctionNames():Array<string> {
//     var ret = [];
//     for (var i = 0; i < methodFunctions.length; i++) {
//         ret.push(methodFunctions[i].name);
//     }
//     return ret;
// }
// ---- Entity ----
function getEntityClassByName(className) {
    return exports.entityClasses[className];
}
exports.getEntityClassByName = getEntityClassByName;
// public static getCollectionClasses():Array<TypeClass<Object>> {
//     var result:Array<TypeClass<Object>> = [];
//     for (var i in entityClasses) {
//         var entityClass = entityClasses[i];
//         if (PersistenceAnnotation.getCollectionName(entityClass))
//             result.push(entityClass);
//     }
//     return result;
// }
function getEntityClasses() {
    var result = [];
    for (var i in exports.entityClasses) {
        var entityClass = exports.entityClasses[i];
        result.push(entityClass);
    }
    return result;
}
exports.getEntityClasses = getEntityClasses;
// static getCollectionName(f:TypeClass<any>):string {
//     return getMetadata("persistence:collectionName", f);
// }
// static isRootEntity(f:TypeClass<any>):boolean {
//     return !!PersistenceAnnotation.getCollectionName(f);
// }
function isEntity(f) {
    return !!exports.entityClasses[getClassName(f)];
}
exports.isEntity = isEntity;
function getDocumentPropertyName(typeClass, objectPropertyName) {
    if (!typeClass)
        return undefined;
    var documentNames = getMetadata("documentNames", typeClass.prototype);
    return documentNames ? documentNames[objectPropertyName] : undefined;
}
exports.getDocumentPropertyName = getDocumentPropertyName;
function getObjectPropertyName(typeClass, documentPropertyName) {
    if (!typeClass)
        return undefined;
    var objectNames = getMetadata("objectNames", typeClass.prototype);
    return objectNames ? objectNames[documentPropertyName] : undefined;
}
exports.getObjectPropertyName = getObjectPropertyName;
function isArrayOrMap(f, propertyName) {
    while (f != Object) {
        if (getPropertyProperty(f, propertyName, "arrayOrMap"))
            return true;
        f = getParentClass(f);
    }
    return false;
}
exports.isArrayOrMap = isArrayOrMap;
// ---- typed properties ----
function getPropertyClass(f, propertyName) {
    while (f != Object) {
        var classNameOfF = getPropertyProperty(f, propertyName, "type");
        if (classNameOfF) {
            var p = getEntityClassByName(classNameOfF);
            if (!p)
                throw new Error('Class ' + f + "', property '" + propertyName + "': Defined as type '" + classNameOfF + "'. Could not find an entity with that name.");
            else
                return p;
        }
        f = getParentClass(f);
    }
    return undefined;
}
exports.getPropertyClass = getPropertyClass;
function getTypedPropertyNames(f) {
    var result = [];
    while (f != Object) {
        var props = getMetadata("property_properties", f);
        for (var i in props) {
            if (props[i].type)
                result.push(i);
        }
        f = getParentClass(f);
    }
    return result;
}
exports.getTypedPropertyNames = getTypedPropertyNames;
function setPropertyProperty(cls, propertyName, property, value) {
    var arr = getMetadata("property_properties", cls.constructor);
    if (!arr) {
        arr = {};
        defineMetadata("property_properties", arr, cls.constructor);
    }
    var propProps = arr[propertyName];
    if (!propProps) {
        propProps = {};
        arr[propertyName] = propProps;
    }
    propProps[property] = value;
}
exports.setPropertyProperty = setPropertyProperty;
function getPropertyNamesOfPropertiesThatHaveProperties(cls) {
    return Object.keys(getMetadata("property_properties", cls));
}
exports.getPropertyNamesOfPropertiesThatHaveProperties = getPropertyNamesOfPropertiesThatHaveProperties;
// this is i.e. good to find all properties on a class that have a "type" property
function getPropertyNamesOfPropertiesThatHaveAProperty(cls, propertyPropertyName) {
    var r = [];
    var props = getMetadata("property_properties", cls);
    for (var i in props) {
        if (props[i][propertyPropertyName]) {
            r.push(i);
        }
    }
    return r;
}
exports.getPropertyNamesOfPropertiesThatHaveAProperty = getPropertyNamesOfPropertiesThatHaveAProperty;
function getPropertyProperty(cls, propertyName, propertyProperty) {
    var arr = getMetadata("property_properties", cls);
    if (arr && arr[propertyName]) {
        return arr[propertyName][propertyProperty];
    }
    return undefined;
}
exports.getPropertyProperty = getPropertyProperty;
function getParentClass(t) {
    if (!t)
        return undefined;
    return Object.getPrototypeOf(t.prototype).constructor;
}
exports.getParentClass = getParentClass;
function getIdPropertyName(t) {
    return getObjectPropertyName(t, "_id") || "_id";
}
exports.getIdPropertyName = getIdPropertyName;
// ---- AsForeignKeys ----
// static isStoredAsForeignKeys(f:TypeClass<any>, propertyName:string):boolean {
//     while (f != Object) {
//         if (PersistenceAnnotation.getPropertyProperty(f, propertyName, "askeys"))
//             return true;
//         f = PersistenceAnnotation.getParentClass(f);
//
//     }
//     return false;
// }
function isIgnored(f, propertyName) {
    //return PersistenceAnnotation.getPropertyProperty(typeClass, propertyName, "ignore");
    while (f != Object) {
        if (getPropertyProperty(f, propertyName, "ignore"))
            return true;
        f = getParentClass(f);
    }
    return false;
}
exports.isIgnored = isIgnored;
function isPrivateToServer(f, propertyName) {
    while (f != Object) {
        if (getPropertyProperty(f, propertyName, "privateToServer"))
            return true;
        f = getParentClass(f);
    }
    return false;
}
exports.isPrivateToServer = isPrivateToServer;
function isParent(f, propertyName) {
    //return PersistenceAnnotation.getPropertyProperty(typeClass, propertyName, "ignore");
    while (f != Object) {
        if (getPropertyProperty(f, propertyName, "parent"))
            return true;
        f = getParentClass(f);
    }
    return false;
}
exports.isParent = isParent;
function getParentPropertyNames(f) {
    return getPropertyNamesOfPropertiesThatHaveAProperty(f, 'parent');
}
exports.getParentPropertyNames = getParentPropertyNames;
// ---- Wrap ----
function getWrappedFunctionNames(f) {
    return getPropertyNamesByMetaData(f.prototype, "persistence:wrap");
}
exports.getWrappedFunctionNames = getWrappedFunctionNames;
function getCollectionUpdateOptions(cls, functionName) {
    if (!cls)
        return undefined;
    return getPropertyProperty(cls.prototype, functionName, "collectionUpdate");
}
exports.getCollectionUpdateOptions = getCollectionUpdateOptions;
function getCollectionUpdateFunctionNames(f) {
    return getPropertyNamesOfPropertiesThatHaveAProperty(f, 'collectionUpdate');
}
exports.getCollectionUpdateFunctionNames = getCollectionUpdateFunctionNames;
function getPropertyNamesByMetaData(o, metaData) {
    var result = [];
    for (var i in o) {
        var value = o[i];
        if (typeof value == "function" && getMetadata(metaData, value))
            result.push(i);
    }
    return result;
}
exports.getPropertyNamesByMetaData = getPropertyNamesByMetaData;
/**
 * @hidden
 */
function monkeyPatch(object, functionName, patchFunction) {
    var originalFunction = object[functionName];
    object[functionName] = function monkeyPatchFunction() {
        var args = [];
        args.push(originalFunction);
        for (var i in arguments) {
            args.push(arguments[i]);
        }
        return patchFunction.apply(this, args);
    };
    object[functionName].originalFunction = originalFunction;
}
/**
 * @hidden
 */
(function () {
    var data;
    if (typeof global != "undefined") {
        if (!global["_reob_data"])
            global["_reob_data"] = {};
        data = global["_reob_data"];
    }
    else if (typeof window != "undefined") {
        if (!window["_reob_data"])
            window["_reob_data"] = {};
        data = window["_reob_data"];
    }
    else
        data = {};
    if (!data.entityClasses)
        data.entityClasses = {};
    exports.entityClasses = data.entityClasses;
})();
//# sourceMappingURL=Annotations.js.map