"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const reob_1 = require("./reob");
const Internal_1 = require("./Internal");
const Annotations_1 = require("./Annotations");
class SerializationPath {
    // this is used when lazy loading properties
    // private objectRetriever:ObjectRetriever;
    constructor(collectionName, id) {
        this.path = collectionName;
        // this.objectRetriever = objectRetriever;
        if (id)
            this.path += "[" + id + "]";
        if (!this.getId()) {
            throw new Error("the id parameter is missing");
        }
    }
    clone() {
        var sp = new SerializationPath(this.path);
        return sp;
    }
    getCollectionName() {
        return this.path.split("[")[0];
    }
    getId() {
        var i1 = this.path.indexOf("[");
        var i2 = this.path.indexOf("]");
        if (i1 != -1 && i2 != -1 && i1 < i2)
            return this.path.split("[")[1].split("]")[0];
        else
            return undefined;
    }
    forEachPathEntry(iterator) {
        if (this.path.indexOf(".") != -1)
            this.path.split("].")[1].split(".").forEach(function (entry) {
                var propertyName = entry;
                var index = undefined;
                if (entry.indexOf("|") != -1) {
                    propertyName = entry.split("|")[0];
                    index = entry.split("|")[1];
                }
                iterator(propertyName, index);
            });
    }
    getSubObject(rootObject) {
        var o = rootObject;
        if (this.path.indexOf(".") != -1) {
            this.path.split("].")[1].split(".").forEach(function (entry) {
                if (o) {
                    if (entry.indexOf("|") != -1) {
                        var p = entry.split("|");
                        var arrayOrMap = o[p[0]];
                        var id = p[1];
                        var foundEntry = false;
                        for (var j in arrayOrMap) {
                            var arrayEntry = arrayOrMap[j];
                            if (arrayEntry.getId && arrayEntry.getId() == id) {
                                o = arrayEntry;
                                foundEntry = true;
                                break;
                            }
                        }
                        if (!foundEntry) {
                            if (arrayOrMap[id])
                                o = arrayOrMap[id];
                            else
                                o = undefined;
                        }
                    }
                    else
                        o = o[entry];
                }
            });
        }
        return o;
    }
    appendArrayOrMapLookup(name, id) {
        this.path += "." + name + "|" + id;
    }
    appendPropertyLookup(name) {
        this.path += "." + name;
    }
    toString() {
        return this.path;
    }
    static setObjectContext(o, sp, handler, request) {
        let object = o;
        if (object) {
            Internal_1.setNonEnumerableProperty(object, "_reobObjectContext", {
                serializationPath: sp,
                handler: handler,
                request: request
            });
        }
    }
    static getObjectContext(o) {
        let object = o;
        return object ? object._reobObjectContext : undefined;
    }
    // if I could I would make this package protected
    static updateObjectContexts(object, handler, request, visited) {
        var that = this;
        if (!visited)
            visited = [];
        if (visited.indexOf(object) != -1)
            return;
        if (!object || typeof object != "object")
            return;
        visited.push(object);
        var objectClass = Annotations_1.getClass(object);
        // // if (PersistenceAnnotation.PersistenceAnnotation.isRootEntity(objectClass)) {
        //     if (!object._ommObjectContext || !object._ommObjectContext.serializationPath) {
        //         var id = PersistenceAnnotation.getId(object);
        //         if ( id ){
        //             var sp = new SerializationPath( PersistenceAnnotation.PersistenceAnnotation.getCollectionName(objectClass), id);
        //             SerializationPath.setObjectContext( object, sp, handler );
        //         }
        //     }
        // // }
        if (!SerializationPath.getObjectContext(object)) {
            // throw new Error("Cant update object context, as there is no object context. Class:"+PersistenceAnnotation.className(objectClass)+" Id:"+PersistenceAnnotation.getId(object));
            return; // we're done here
        }
        Annotations_1.getTypedPropertyNames(objectClass).forEach(function (typedPropertyName) {
            // if (!PersistenceAnnotation.isStoredAsForeignKeys(objectClass, typedPropertyName)) {
            //     //console.log("updating foreignkey property " + typedPropertyName);
            var v = object[typedPropertyName];
            if (v) {
                if (Annotations_1.isArrayOrMap(objectClass, typedPropertyName)) {
                    //console.log("updating foreignkey property " + typedPropertyName + " is array");
                    for (var i in v) {
                        var e = v[i];
                        if (e) {
                            var index = i;
                            if (reob_1.getId(e))
                                index = reob_1.getId(e);
                            //console.log("updating persistnece path for isArrayOrMap " + typedPropertyName + "  key:" + i + " value:", e, "object: ", object);
                            var sp = SerializationPath.getObjectContext(object).serializationPath.clone();
                            sp.appendArrayOrMapLookup(typedPropertyName, index);
                            SerializationPath.setObjectContext(e, sp, handler, request);
                            that.updateObjectContexts(e, handler, request, visited);
                        }
                    }
                }
                else {
                    //console.log("updating foreignkey property direct property " + typedPropertyName);
                    var sp = SerializationPath.getObjectContext(object).serializationPath.clone();
                    sp.appendPropertyLookup(typedPropertyName);
                    SerializationPath.setObjectContext(v, sp, handler, request);
                }
            }
        });
    }
}
exports.SerializationPath = SerializationPath;
//# sourceMappingURL=SerializationPath.js.map