"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Cloner = require("./Cloner");
const SubObjectPath_1 = require("./SubObjectPath");
const reob_1 = require("./reob");
const Annotations_1 = require("./Annotations");
class Serializer {
    constructor() {
    }
    static forEachTypedObject(object, cb) {
        this.forEachTypedObjectRecursive(object, object, new SubObjectPath_1.SubObjectPath(), [], cb);
    }
    static forEachTypedObjectRecursive(rootObject, object, path, visited, cb) {
        var that = this;
        if (visited.indexOf(object) != -1)
            return;
        if (!object || typeof object != "object")
            return;
        visited.push(object);
        cb(path, object);
        var objectClass = Annotations_1.getClass(object);
        Annotations_1.getTypedPropertyNames(objectClass).forEach(function (typedPropertyName) {
            if (!Annotations_1.isParent(objectClass, typedPropertyName)) {
                //console.log("updating foreignkey property " + typedPropertyName);
                var v = object[typedPropertyName];
                if (v) {
                    if (Annotations_1.isArrayOrMap(objectClass, typedPropertyName)) {
                        //console.log("updating foreignkey property " + typedPropertyName + " is array");
                        for (var i in v) {
                            var e = v[i];
                            //console.log("updating persistnece path for isArrayOrMap " + typedPropertyName + "  key:" + i + " value:", e, "object: ", object);
                            var subObjectPath = path.clone();
                            if (typeof reob_1.getId(e) != "undefined") {
                                subObjectPath.appendArrayOrMapLookup(typedPropertyName, reob_1.getId(e));
                            }
                            else {
                                subObjectPath.appendArrayOrMapLookup(typedPropertyName, i);
                            }
                            cb(subObjectPath, e);
                            that.forEachTypedObjectRecursive(rootObject, e, subObjectPath, visited, cb);
                        }
                    }
                    else {
                        var subObjectPath = path.clone();
                        subObjectPath.appendPropertyLookup(typedPropertyName);
                        cb(subObjectPath, v);
                        that.forEachTypedObjectRecursive(rootObject, v, subObjectPath, visited, cb);
                    }
                }
            }
        });
    }
    toObject(doc, handler, f, serializationPath, request) {
        var o;
        if (Array.isArray(doc)) {
            var r = [];
            for (var j = 0; j < doc.length; j++) {
                r[j] = this.toObject(doc[j], handler, f);
            }
            o = r;
        }
        else if (!doc || typeof doc == "string" || typeof doc == "number" || typeof doc == "boolean")
            o = doc;
        else
            o = this.toObjectRecursive(doc, undefined, undefined, f, handler, request);
        if (handler && serializationPath) {
            reob_1.SerializationPath.setObjectContext(o, serializationPath, handler, request);
            reob_1.SerializationPath.updateObjectContexts(o, handler, request);
        }
        return o;
    }
    toObjectRecursive(doc, parent, parentPropertyName, f, handler, request) {
        var o;
        if (!doc)
            return doc;
        if (typeof doc == "function")
            throw new Error("Error in 'toObject'. doc is a function.");
        if (f && typeof f["toObject"] == "function") {
            //console.log("using the custom toObject function of class "+omm.className(f));
            o = f["toObject"](doc, this, parent, parentPropertyName);
        }
        else {
            // if the document contains a property called "className" it defines the class that's going to be instantiated
            if (doc._className) {
                f = Annotations_1.getEntityClassByName(doc._className);
            }
            if (!f)
                return Cloner.clone(doc);
            //     throw new Error("Could not determine class of document. Either the document needs to have a '_className' property or a class needs to be passed to the serializer. Document: "+ JSON.stringify( doc ) );
            // instantiate the new object
            if (f && f.prototype) {
                o = Object.create(f.prototype);
            }
            else
                o = {};
            if (doc._serializationPath) {
                var sp = new reob_1.SerializationPath(doc._serializationPath);
                reob_1.SerializationPath.setObjectContext(o, sp, handler, request);
            }
            Annotations_1.getParentPropertyNames(f).forEach(function (parentPropertyName) {
                o[parentPropertyName] = parent;
            });
            // iterate over all properties
            for (var propertyName in doc) {
                if (propertyName == "_className" || propertyName == "_serializationPath")
                    continue;
                var value = doc[propertyName];
                var objectNameOfTheProperty = f ? Annotations_1.getObjectPropertyName(f, propertyName) : undefined;
                if (!objectNameOfTheProperty)
                    objectNameOfTheProperty = propertyName;
                var propertyClass = Annotations_1.getPropertyClass(f, objectNameOfTheProperty);
                // var isStoredAsKeys = PersistenceAnnotation.isStoredAsForeignKeys(f, objectNameOfTheProperty);
                if (propertyClass /*&& !isStoredAsKeys*/) {
                    if (Annotations_1.isArrayOrMap(f, objectNameOfTheProperty)) {
                        var result = Array.isArray(value) ? [] : {};
                        for (var i in value) {
                            var entry = value[i];
                            entry = this.toObjectRecursive(entry, o, propertyName, propertyClass, handler, request);
                            result[i] = entry;
                        }
                        // this can only happen once because if the property is accessed the "lazy load" already kicks in
                        o[objectNameOfTheProperty] = result;
                    }
                    else {
                        o[objectNameOfTheProperty] = this.toObjectRecursive(value, o, propertyName, propertyClass, handler, request);
                    }
                }
                else {
                    o[objectNameOfTheProperty] = Cloner.clone(value);
                    // o[objectNameOfTheProperty] = value;
                }
            }
            // call "postCreate" functions
            let postCreateFunctionNames = Annotations_1.getPostCreateFunctionNames(f);
            if (postCreateFunctionNames) {
                postCreateFunctionNames.forEach(functionName => {
                    o[functionName]();
                });
            }
        }
        return o;
    }
    // this is the function that gets called when the process of converting an object to a document starts.
    // its called once in the process
    toDocument(object, includeContext, omitPropertiesPrivateToServer) {
        const objectClass = Annotations_1.getClass(object);
        const result = this.toDocumentRecursive(object, objectClass, omitPropertiesPrivateToServer);
        // console.log("returning document:",result);
        return result;
    }
    toDocumentRecursive(object, expectedClass, includeContext, omitPropertiesPrivateToServer) {
        var result;
        var objectClass = Annotations_1.getClass(object);
        var context = reob_1.SerializationPath.getObjectContext(object);
        // console.log(" documenting ", object,  typeof object);
        // is there a customized way to create a document ?
        if (objectClass && typeof objectClass.toDocument == "function") {
            result = objectClass.toDocument(object, this, includeContext, omitPropertiesPrivateToServer);
        }
        else if (!object || typeof object == "string" || typeof object == "number" || typeof object == "boolean") {
            result = object;
        }
        else if (Array.isArray(object) || Annotations_1.isArrayOrMap(objectClass, property)) {
            result = Array.isArray(object) ? [] : {};
            const expectedClass = Annotations_1.getPropertyClass(objectClass, property);
            for (var property in object) {
                const subDoc = this.toDocumentRecursive(object[property], expectedClass, includeContext, omitPropertiesPrivateToServer);
                result[property] = subDoc;
            }
        }
        else if (typeof object == "object") {
            result = {};
            // for all enumerable properties in the object ...
            for (var property in object) {
                const value = object[property];
                const expectedClass = Annotations_1.getPropertyClass(objectClass, property);
                // figure out if the property needs to be part of the document or not
                let canBeIgnored = omitPropertiesPrivateToServer && Annotations_1.isPrivateToServer(objectClass, property); // its private to the server
                canBeIgnored = canBeIgnored || Annotations_1.isIgnored(objectClass, property);
                canBeIgnored = canBeIgnored || value === undefined; // it's undefined
                canBeIgnored = canBeIgnored || Annotations_1.isParent(objectClass, property); // parent properties are not part of documents, they're deducted
                canBeIgnored = canBeIgnored || typeof value === "function"; // its's a function
                if (!canBeIgnored) {
                    let documentNameOfTheProperty = Annotations_1.getDocumentPropertyName(objectClass, property) || property;
                    const subDoc = this.toDocumentRecursive(value, expectedClass, includeContext, omitPropertiesPrivateToServer);
                    result[documentNameOfTheProperty] = subDoc;
                }
            }
            // add the serialization path if needed (for network)
            if (includeContext && context && context.serializationPath) {
                result['_serializationPath'] = context.serializationPath.toString();
            }
            // add the object class if it deviates from what's expected or its required
            if (objectClass && Annotations_1.isEntity(objectClass) && (includeContext || expectedClass != objectClass)) {
                result['_className'] = Annotations_1.getClassName(objectClass);
            }
        }
        else if (typeof object == "function") {
            // ignore
        }
        else {
            throw new Error("Unexpected object type :" + (typeof object));
        }
        // console.log(" documenting result:", result);
        return result;
    }
}
exports.Serializer = Serializer;
//# sourceMappingURL=Serializer.js.map