"use strict";
/**
 * Created by bert on 07.07.16.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const reob = require("./serverModule");
const wm = require("web-methods");
const expressModule = require("express");
const http = require("http");
const path = require("path");
const fs = require("fs");
const compression = require("compression");
const mongodb_1 = require("mongodb");
class Server {
    constructor(theMongoDbOrDbUrl) {
        this.collections = {};
        this.collectionLoadingEnabled = {};
        this.services = {};
        this.beforeMethodListener = [];
        this.afterMethodListener = [];
        this.webMethodsAdded = [];
        this.webMethods = new wm.WebMethods();
        this.serializer = new reob.Serializer();
        this.express = expressModule();
        this.express.use(compression());
        this.httpServer = http.createServer(this.express);
        if (typeof theMongoDbOrDbUrl == "string") {
            this.mongoUrl = theMongoDbOrDbUrl;
        }
        else {
            this.mongoDb = theMongoDbOrDbUrl;
        }
    }
    getWebRootPath() {
        return this.webRootPath;
    }
    stop() {
        this.httpServer.close();
    }
    connectToMongoDb() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.mongoDb) {
                //let client = new MongoClient(this.mongoUrl);
                let client = yield mongodb_1.MongoClient.connect(this.mongoUrl);
                this.mongoDb = client.db();
            }
            for (var i in this.collections) {
                var c = this.collections[i];
                c.setMongoCollection(this.mongoDb);
            }
        });
    }
    start(port) {
        return __awaiter(this, void 0, void 0, function* () {
            this.webMethods.registerEndpoint(this.express);
            this.addAllEntityWebMethods();
            this.registerGetter();
            if (this.webRootPath) {
                this.registerStaticGET();
            }
            this.port = port || 8080;
            yield this.connectToMongoDb();
            yield new Promise((resolve, reject) => {
                this.httpServer.listen(this.port, () => {
                    resolve();
                });
            });
        });
    }
    getExpress() {
        return this.express;
    }
    serveStatic(theWebRootPath, indexFileName) {
        this.webRootPath = path.resolve(theWebRootPath);
        this.indexFileName = indexFileName ? indexFileName : "index.html";
    }
    registerStaticGET() {
        if (this.webRootPath && this.indexFileName) {
            this.express.get('/*', (req, res, next) => {
                //This is the current file they have requested
                var file = req.params[0];
                // if( file.indexOf( "js/")!=0 && file.indexOf( "css/")!=0 && file.indexOf( "fonts/")!=0  && file.indexOf( "img/" )!=0 && file.indexOf( "src/")!=0 &&file.indexOf( "components/")!=0 &&file.indexOf( "bootstrap/")!=0 && file.indexOf( "bundle")!=0 )
                //     file = indexFileName;
                var fileName = path.resolve(this.webRootPath, file);
                fs.exists(fileName, (exists) => {
                    if (!exists)
                        fileName = path.resolve(this.webRootPath, this.indexFileName);
                    res.sendFile(fileName);
                });
            });
            this.express.get('/', (req, res, next) => {
                var fileName = path.resolve(this.webRootPath, this.indexFileName);
                res.sendFile(fileName);
            });
        }
    }
    addCollection(c) {
        this.collections[c.getName()] = c;
        if (this.mongoDb)
            c.setMongoCollection(this.mongoDb);
    }
    setLoadingAllowed(c, i) {
        this.collectionLoadingEnabled[c.getName()] = i;
    }
    addService(serviceName, service) {
        // singletons dont need a
        var o;
        if (typeof service == "object") {
            reob.SerializationPath.setObjectContext(service, undefined, this, undefined);
            o = service;
        }
        else {
            o = service({});
        }
        var serviceClass = reob.getClass(o);
        // var serviceName = omm.getServiceName(serviceClass);
        // if( !serviceName )
        //     throw new Error("Added service has no @Service decorator. Class name:"+ omm.getClassName(o) );
        this.services[serviceName] = service;
        var serviceClass = reob.getClass(o);
        reob.getRemoteFunctionNames(serviceClass).forEach((remoteFunctionName) => {
            var options = reob.getMethodOptions(serviceClass, remoteFunctionName);
            if (!options.name) {
                options.name = serviceName + "." + options.propertyName;
            }
            this.addWebMethod(options);
        });
    }
    notifyBeforeMethodListeners(object, functionName, args, request) {
        var promises = [];
        this.beforeMethodListener.forEach((ml) => {
            promises.push(reob.EventListenerRegistry.castToPromise(() => { ml(object, functionName, args, request); }));
        });
        return Promise.all(promises).then(() => { });
    }
    notifyAfterMethodListeners(object, functionName, args, request, result) {
        var promises = [];
        this.afterMethodListener.forEach((ml) => {
            promises.push(reob.EventListenerRegistry.castToPromise(() => { ml(object, functionName, args, request, result); }));
        });
        return Promise.all(promises).then(() => { });
    }
    onBeforeMethod(eventHandler) {
        this.beforeMethodListener.push(eventHandler);
    }
    onAfterMethod(eventHandler) {
        this.afterMethodListener.push(eventHandler);
    }
    removeAllMethodListeners() {
        this.beforeMethodListener = [];
        this.afterMethodListener = [];
    }
    addAllEntityWebMethods() {
        reob.getEntityClasses().forEach((c) => {
            reob.getRemoteFunctionNames(c).forEach((remoteFunctionName) => {
                var options = reob.getMethodOptions(c, remoteFunctionName);
                if (!options.name) {
                    options.name = reob.getClassName(c) + "." + options.propertyName;
                }
                this.addWebMethod(options);
            });
        });
    }
    getPort() {
        return this.port;
    }
    setRequestFactory(f) {
        this.requestFactory = f;
    }
    createRequest(userData) {
        if (this.requestFactory) {
            return this.requestFactory(userData);
        }
        else {
            return { userData };
        }
    }
    addWebMethod(options) {
        if (this.webMethodsAdded.indexOf(options.name) != -1)
            return;
        this.webMethodsAdded.push(options.name);
        this.webMethods.add(options.name, (...args) => {
            var startTime = Date.now();
            // the object id is the first parameter
            var objectId = args.shift();
            // the user Data is the second parameter
            var userData = args.shift();
            var request = this.createRequest(userData);
            // convert parameters given to the web method from documents to objects
            this.convertWebMethodParameters(args, options.parameterTypes, userData);
            // load object based on the object id. this could either be a registered object or an object form a collection
            var objectPromise = this.retrieveObject(objectId, request).then((object) => {
                // this might be the collection update or another function that is called directly
                return this.notifyBeforeMethodListeners(object, options.name.split(".")[1], args, request).then(() => {
                    return object;
                });
            });
            var resultPromise = objectPromise.then((object) => {
                return object[options.propertyName].apply(object, args);
            });
            return Promise.all([objectPromise, resultPromise]).then((values) => {
                var object = values[0];
                var result = values[1];
                var res = {};
                if (result) {
                    res.document = this.serializer.toDocument(result, true, true);
                }
                return this.notifyAfterMethodListeners(object, options.name.split(".")[1], args, request, result).then(() => {
                    return res;
                });
            });
        });
    }
    getCollection(objectId) {
        var sPath = new reob.SerializationPath(objectId);
        var collectionName = sPath.getCollectionName();
        var collection = collectionName ? this.collections[collectionName] : undefined;
        return collection;
    }
    retrieveObject(objectId, request) {
        var service = this.services[objectId];
        if (service) {
            var s = service;
            if (typeof service == "function") {
                s = service(request);
            }
            return Promise.resolve(s);
        }
        else {
            if (typeof objectId != "string")
                throw new Error("Path needs to be a string");
            var sPath = new reob.SerializationPath(objectId);
            var collection = this.getCollection(objectId);
            if (collection) {
                return collection.getById(sPath.getId(), request).then((o) => {
                    return sPath.getSubObject(o);
                });
            }
            else
                return Promise.reject("No collection found to retrieve object. Key:" + objectId);
        }
    }
    registerGetter() {
        this.webMethods.add("get", (collectionName, objectId, userData) => {
            var request = this.createRequest(userData);
            var allowedPromise;
            return new Promise((resolve, reject) => {
                if (collectionName)
                    resolve(this.retrieveObject(collectionName + "[" + objectId + "]", request));
                else
                    reject(new Error("No collection name given"));
            }).then((object) => {
                var i = this.collectionLoadingEnabled[collectionName];
                if (typeof i == "function") {
                    return Promise.resolve(i(objectId, request, object)).then(() => object);
                }
                else {
                    return Promise.resolve(i).then(() => object);
                }
            }).then((obj) => {
                var doc = obj ? this.serializer.toDocument(obj, true, true) : undefined;
                return doc;
            });
        });
    }
    // converts parameters given to the web method from documents to objects
    convertWebMethodParameters(args, classNames, request) {
        for (var i = 0; i < args.length; i++) {
            if (classNames && classNames.length > i) {
                var cls = reob.getEntityClassByName(classNames[i]);
                if (cls) {
                    if (typeof args[i] == "string")
                        args[i] = this.retrieveObject(args[i], request);
                    else if (typeof args[i] == "object")
                        args[i] = this.serializer.toObject(args[i], cls);
                }
            }
        }
    }
}
exports.Server = Server;
//# sourceMappingURL=Server.js.map